unit CustomFileSystem;

interface

uses
  SysUtils, Classes, Windows, ShellAPI, Graphics, Forms, Controls;

type
  TFileDescriptionRecord=record
    FileType:string;
    OpenImageIndex,CloseImageIndex:Integer;
  end;

  PFileDescriptionRecord=^TFileDescriptionRecord;

  TCharSet=set of Char;

  TFileItemDescriptor=record
    FileName,FilePath,FileTypeName:string;
    Folder,Readable,Writable,FastExplore:Boolean;
    FileSize:Int64;
    DateModified:TDateTime;
    OpenShellImageIndex,CloseShellImageIndex:Integer;
  end;

  PFileItemDescriptor=^TFileItemDescriptor;

  TFileItemList=class
  private
    FList:TList;

    function GetCount:Integer;
    function GetItems(x: Integer): TFileItemDescriptor;

    procedure Clear;
  public
    constructor Create;

    procedure AddItem(f:TFileItemDescriptor);
    procedure DeleteItem(x:Integer);

    property Count:Integer read GetCount;
    property Items[x:Integer]:TFileItemDescriptor read GetItems;

    destructor Destroy;override;
  end;

  TCustomFileSystem=class;

  IFileSystemHook=interface
    ['{A949B9ED-9DBF-4F4A-AEC3-DE9DC66AA899}']
    function GetImplementationComponent:TComponent;
    property ImplementationComponent:TComponent read GetImplementationComponent;

    function GetCurrentDirectory:string;
    procedure SetCurrentDirectory(const Value:string);
    property CurrentDirectory:string read GetCurrentDirectory write SetCurrentDirectory;

    procedure SetFileSystem(const Value:TCustomFileSystem);

    procedure DoConnect;
    procedure DoDisconnect;
    procedure DoUpdate;
  end;

  TCustomFileSystemHook=class(TComponent,IFileSystemHook)
  private
    FFileSystem:TCustomFileSystem;
    FCOmponent:TComponent;

    procedure SetCurrentDirectory(const Value:string);virtual;abstract;

    procedure SetFileSystem(const Value: TCustomFileSystem);

    procedure DoConnect;virtual;abstract;
    procedure DoDisconnect;virtual;abstract;
    procedure DoUpdate;virtual;abstract;
  protected
    function GetImplementationComponent:TComponent;virtual;

    function GetCurrentDirectory:string;virtual;
  public
    constructor Create(AOwner:TComponent);override;

    destructor Destroy;override;
  published
    property FileSystem:TCustomFileSystem read FFileSystem write SetFileSystem;
  end;

  THookType=(hConnect,hUpdate,hCurrentDirectory,hDisconnect);

  TCustomFileSystem=class(TComponent)
  private
    function GetCurrentDirectory:string;
    procedure SetCurrentDirectory(Value: string);virtual;
  protected
    FCurrentDirectory:string;
    FHooks:TList;
    FRootDescription:TFileItemDescriptor;

    function GetConnected:Boolean;virtual;abstract;
    procedure NotifyHooks(AHookType:THookType);
  public
    constructor Create(AOwner:TComponent);override;

    property RootRecordDescription:TFileItemDescriptor read FRootDescription;

    procedure Hook(AHook:IFileSystemHook);virtual;
    procedure UnHook(AHook:IFileSystemHook);virtual;

    property Connected:Boolean read GetConnected;
    procedure Connect;virtual;abstract;
    procedure Disconnect(Force:Boolean=False);virtual;
    procedure UpdateHooks;virtual;

    function GetDirectoryContent(Path:string):TFileItemList;virtual;
    function DirectoryExists(Path:string):Boolean;virtual;abstract;
    procedure CreateDirectory(Path:string);virtual;abstract;
    procedure DeleteDirectory(Path:string);virtual;abstract;
    procedure RenameDirectory(OldPath,NewPath:string);virtual;abstract;
    procedure DirectoryUp;virtual;

    function FileExists(FileName:string):Boolean;virtual;abstract;
    procedure RenameFile(OldFileName,NewFileName:string);virtual;abstract;
    procedure DeleteFile(FileName:string);virtual;abstract;
    procedure WriteStreamToFile(FileName:string;Stream:TStream);virtual;abstract;
    procedure ReadStreamFromFile(FileName:string;Stream:TStream);virtual;abstract;

    property CurrentDirectory:string read GetCurrentDirectory write SetCurrentDirectory;

    destructor Destroy;override;
  end;

procedure SplitString(s:string;var SLeft,SRight:string;Separators:TCharSet;FromRight:Boolean=True);
procedure GetFileExtensionShellDescription(Ext:string;var FileTypeName:string;var ShellImageIndex:Integer);

var
  ShellSmallImageList,ShellLargeImageList:TImageList;
  RegisteredExtensions:TStringList;
  UpArrowShellImageIndex,DownArrowShellImageIndex:Integer;
  DefaultFolderDescription:TFileDescriptionRecord;

implementation

procedure SplitString(s:string;var SLeft,SRight:string;Separators:TCharSet;FromRight:Boolean=True);
var
  a:Integer;
begin
  if s='' then begin
    SLeft:='';
    SRight:='';
    Exit;
  end;
  if FromRight then begin
    a:=Length(s);
    while (a>0) and not (s[a] in Separators) do
      Dec(a);
    if a=0 then begin
      SLeft:='';
      SRight:=s;
    end else begin
      SLeft:=Copy(s,1,a-1);
      SRight:=Copy(s,a+1,Length(s)-a);
    end;
  end else begin
    a:=0;
    while (a<=Length(s)) and not (s[a] in Separators) do
      Inc(a);
    if a>Length(s) then begin
      SLeft:=s;
      SRight:='';
    end else begin
      SLeft:=Copy(s,1,a-1);
      SRight:=Copy(s,a+1,Length(s)-a);
    end;
  end;
end;

procedure GetFileExtensionShellDescription(Ext:string;var FileTypeName:string;var ShellImageIndex:Integer);
var
  a,b:Integer;
  t:string;
  l:TImageList;
  sfi:TSHFileInfo;
  i:TIcon;
  q:PChar;
  p:PFileDescriptionRecord;
begin
  FileTypeName:='';
  a:=RegisteredExtensions.IndexOf(Ext);
  if a=-1 then begin
    New(p);
    p^.FileType:='';
    i:=TIcon.Create;
    for b:=0 to 1 do begin
      if b=0 then
        l:=ShellSmallImageList
      else
        l:=ShellLargeImageList;
      GetMem(q,MAX_PATH);
      GetTempPath(MAX_PATH,q);
      t:=q+'TEMP'+Ext;
      FreeMem(q);
        TFileStream.Create(t,fmCreate or fmOpenWrite).Destroy;
      if SHGetFileInfo(PChar(t),0,sfi,SizeOf(sfi),SHGFI_TYPENAME or SHGFI_ICON or (SHGFI_SMALLICON*(1-b)+b*SHGFI_LARGEICON))<>0 then begin
        i.ReleaseHandle;
        i.Handle:=sfi.hIcon;
        p^.FileType:=sfi.szTypeName;
      end;
      p^.OpenImageIndex:=l.AddIcon(i);
      p^.CloseImageIndex:=p^.OpenImageIndex;
      if b=1 then
        a:=RegisteredExtensions.AddObject(Ext,TObject(p));
    end;
    i.Destroy;
  end;
  with PFileDescriptionRecord(RegisteredExtensions.Objects[a])^ do begin
    ShellImageIndex:=CloseImageIndex;
    FileTypeName:=FileType;
  end;
end;

{ TFileItemList }

procedure TFileItemList.Additem(f: TFileItemDescriptor);
var
  p:PFileItemDescriptor;
begin
  New(p);
  p^:=f;
  FList.Add(p);
end;

procedure TFileItemList.Clear;
var
  a:Integer;
begin
  for a:=FList.Count-1 downto 0 do
    Dispose(FList[a]);
  FList.Clear;
end;

constructor TFileItemList.Create;
begin
  inherited;
  FList:=TList.Create;
end;

procedure TFileItemList.DeleteItem(x: Integer);
begin
  Dispose(FList[x]);
  FList.Delete(x);
end;

destructor TFileItemList.Destroy;
begin
  Clear;
  FList.Destroy;
  inherited;
end;

function TFileItemList.GetCount: Integer;
begin
  Result:=FList.Count;
end;

function TFileItemList.GetItems(x: Integer): TFileItemDescriptor;
begin
  Result:=PFileItemDescriptor(FList.Items[x])^;
end;

{ TCustomFileSystemHook }

constructor TCustomFileSystemHook.Create(AOwner: TComponent);
begin
  inherited;
  FComponent:=AOwner;
end;

destructor TCustomFileSystemHook.Destroy;
begin
  FileSystem:=nil;
  inherited;
end;

function TCustomFileSystemHook.GetCurrentDirectory: string;
begin
  if Assigned(FileSystem) then
    Result:=FileSystem.CurrentDirectory
  else
    Result:='';
end;

function TCustomFileSystemHook.GetImplementationComponent: TComponent;
begin
  Result:=FComponent;
end;

procedure TCustomFileSystemHook.SetFileSystem(const Value: TCustomFileSystem);
begin
  if Value<>FFileSystem then begin
    if Assigned(FFileSystem) then
      FFileSystem.UnHook(FComponent as IFileSystemHook);
    FFileSystem:=Value;
    if Assigned(FFileSystem) then
      FFileSystem.Hook(FComponent as IFileSystemHook);
  end;
end;

{ TCustomFileSystem }

constructor TCustomFileSystem.Create(AOwner: TComponent);
begin
  inherited;
  FHooks:=TList.Create;
  with FRootDescription do begin
    FileName:='';
    FilePath:='';
    FileTypeName:='$ROOT$';
    Folder:=True;
    Readable:=True;
    Writable:=False;
    FastExplore:=True;
    FileSize:=0;
    DateModified:=0;
    OpenShellImageIndex:=DefaultFolderDescription.OpenImageIndex;
    CloseShellImageIndex:=DefaultFolderDescription.CloseImageIndex;
  end;
end;

destructor TCustomFileSystem.Destroy;
var
  a:Integer;
begin
  for a:=FHooks.Count-1 downto 0 do
    IFileSystemHook(FHooks[a]).SetFileSystem(nil);
  FHooks.Destroy;
  inherited;
end;

procedure TCustomFileSystem.DirectoryUp;
var
  s,t:string;
begin
  if CurrentDirectory<>'/' then begin
    SplitString(Copy(CurrentDirectory,1,Length(CurrentDirectory)-1),s,t,['/']);
    CurrentDirectory:=s+'/';
  end;
end;

procedure TCustomFileSystem.Disconnect(Force: Boolean);
begin
  FCurrentDirectory:='';
end;

function TCustomFileSystem.GetCurrentDirectory: string;
begin
  Result:=FCurrentDirectory;
end;

function TCustomFileSystem.GetDirectoryContent(
  Path: string): TFileItemList;
begin
  Result:=TFileItemList.Create;
end;

procedure TCustomFileSystem.Hook(AHook: IFileSystemHook);
begin
  if FHooks.IndexOf(AHook.ImplementationComponent)=-1 then
    FHooks.Add(AHook.ImplementationComponent);
  if Connected then begin
    AHook.DoConnect;
    AHook.CurrentDirectory:=CurrentDirectory;
  end;
end;

procedure TCustomFileSystem.NotifyHooks(AHookType: THookType);
var
  a:Integer;
begin
  for a:=FHooks.Count-1 downto 0 do begin
    with TComponent(FHooks[a]) as IFileSystemHook do
      case AHookType of
        hConnect:DoConnect;
        hUpdate:DoUpdate;
        hCurrentDirectory:CurrentDirectory:=Self.CurrentDirectory;
        hDisconnect:DoDisconnect;
      end;
  end;
end;

procedure TCustomFileSystem.SetCurrentDirectory(Value: string);
begin
  if Value='//' then
    Value:='/';
  if Value<>FCurrentDirectory then begin
    Assert(DirectoryExists(Value),'Folder "'+Value+'" does not exist');
    FCurrentDirectory:=Value;
    NotifyHooks(hCurrentDirectory);
  end;
end;

procedure TCustomFileSystem.UnHook(AHook: IFileSystemHook);
var
  a:Integer;
begin
  a:=FHooks.IndexOf(AHook.ImplementationComponent);
  if a>-1 then 
    FHooks.Remove(AHook.ImplementationComponent);
end;

procedure TCustomFileSystem.UpdateHooks;
begin
  if Connected then
    NotifyHooks(hUpdate);
end;

{$WARNINGS OFF}
procedure OpenFolderShellImages;
var
  i:TIcon;
  p:TBitmap;
  sfi:TSHFileInfo;
  l:TImageList;
  a,j:Integer;
  c:Cardinal;
  _i:PInteger;
begin
  ShellSmallImageList:=TImageList.Create(Application);
  ShellLargeImageList:=TImageList.Create(Application);
  ShellSmallImageList.BkColor:=clWhite;
  ShellLargeImageList.BkColor:=clWhite;
  ShellSmallImageList.BlendColor:=clHighlight;
  ShellLargeImageList.BlendColor:=clHighlight;
  RegisteredExtensions:=TStringList.Create;
  ShellSmallImageList.Width:=16;
  ShellSmallImageList.Height:=16;
  ShellLargeImageList.Width:=32;
  ShellLargeImageList.Height:=32;
  p:=TBitmap.Create;
  for a:=0 to 3 do begin
    _i:=nil;
    case a of
      0,1:l:=ShellSmallImageList;
      2,3:begin
        l:=ShellLargeImageList;
        if a=2 then
          _i:=@UpArrowShellImageIndex
        else
          _i:=@DownArrowShellImageIndex;
      end;
    end;
    p.Width:=l.Width;
    p.Height:=l.Height;
    j:=(a mod 2);
    DrawFrameControl(p.Canvas.Handle,Rect(-2,-2,p.Width+2,p.Height+2),DFC_SCROLL,j*DFCS_SCROLLUP+(1-j)*DFCS_SCROLLDOWN);
    j:=l.Add(p,nil);
    if Assigned(_i) then
      _i^:=j;
  end;
  p.Destroy;
  i:=TIcon.Create;
  for a:=0 to 3 do begin
    case a of
      0,1:begin
        l:=ShellSmallImageList;
        c:=SHGFI_SMALLICON or (a*SHGFI_OPENICON);
        _i:=nil;
      end;
      2,3:begin
        l:=ShellLargeImageList;
        c:=SHGFI_LARGEICON or ((a-2)*SHGFI_OPENICON);
        if a=2 then
          _i:=@DefaultFolderDescription.CloseImageIndex
        else
          _i:=@DefaultFolderDescription.OpenImageIndex;
      end;
    end;
    if SHGetFileInfo(PChar(ExtractFilePath(Application.ExeName)),0,sfi,SizeOf(sfi),SHGFI_ICON or c)<>0 then begin
      i.ReleaseHandle;
      i.Handle:=sfi.hIcon;
      j:=l.AddIcon(i);
    end else
      j:=-1;
    if Assigned(_i) then
      _i^:=j;
  end;
  i.Destroy;
end;
{$WARNINGS ON}

procedure CloseFolderShellImages;
var
  a:Integer;
begin
  for a:=RegisteredExtensions.Count-1 downto 0 do
    Dispose(PFileDescriptionRecord(RegisteredExtensions.Objects[a]));
  RegisteredExtensions.Destroy;
end;

initialization
  OpenFolderShellImages;
finalization
  CloseFolderShellImages;
end.
