unit FTPFileSystem;

interface

uses
  Windows, ShellAPI, SysUtils, Classes, Controls, CustomFileSystem, IdComponent, IdTCPConnection,
  IdTCPClient, IdFTP, IdFTPList, Graphics, SyncObjs, Dialogs, FTPLoginDialog, StrUtils;

type
  TEchoEvent=procedure(Sender:TObject;EchoStr:string) of object;

  TFTPFileSystem = class(TCustomFileSystem)
  private
    FFTP:TidFTP;
    FTransferring:Boolean;
    FOnEcho: TEchoEvent;
    FCriticalSection:TCriticalSection;
    FLoginDialog: TFTPLoginDialog;

    function GetHost: string;
    function GetPassWord: string;
    function GetUserName: string;
    procedure SetHost(const Value: string);
    procedure SetPassWord(const Value: string);
    procedure SetUserName(const Value: string);
    procedure SetOnEcho(const Value: TEchoEvent);

    procedure Status(axSender:TObject;const axStatus:TIdStatus;const asStatusText:String);
    procedure SetLoginDialog(const Value: TFTPLoginDialog);

    procedure Work(Sender:TObject;AWorkMode:TWorkMode;const AWorkCount:Integer);
    procedure WorkBegin(Sender:TObject;AWorkMode:TWorkMode;const AWorkCountMax:Integer);
    procedure WorkEnd(Sender:TObject;AWorkMode:TWorkMode);

    procedure Disconnected(Sender:TObject);
  protected
    function GetConnected:Boolean;override;

    procedure Notification(AComponent:TComponent;Operation:TOperation);override;
  public
    constructor Create(AOwner:TComponent);override;

    procedure Connect;override;
    procedure Disconnect(Force:Boolean=False);override;

    procedure Echo(s:string);

    function GetDirectoryContent(Path:string):TFileItemList;override;

    function DirectoryExists(Path:string):Boolean;override;
    procedure CreateDirectory(Path:string);override;
    procedure DeleteDirectory(Path:string);override;
    procedure RenameDirectory(OldPath,NewPath:string);override;

    function FileExists(FileName:string):Boolean;override;
    procedure RenameFile(OldFileName,NewFileName:string);override;
    procedure DeleteFile(FileName:string);override;
    procedure WriteStreamToFile(FileName:string;Stream:TStream);override;
    procedure ReadStreamFromFile(FileName:string;Stream:TStream);override;

    destructor Destroy;override;
  published
    property Host:string read GetHost write SetHost;
    property UserName:string read GetUserName write SetUserName;
    property PassWord:string read GetPassWord write SetPassWord;

    property LoginDialog:TFTPLoginDialog read FLoginDialog write SetLoginDialog;

    property OnEcho:TEchoEvent read FOnEcho write SetOnEcho;
  end;

implementation

function FTPItemToFileItemDescriptor(Path:string;f:TIdFTPListItem):TFileItemDescriptor;
begin
  with Result do begin
    FileName:=f.FileName;
    if (Path='') then begin
      if FileName='' then
        FilePath:=''
      else
        FilePath:='/';
    end else begin
      if Path[Length(Path)]='/' then
        FilePath:=Path
      else
        FilePath:=Path+'/';
    end;
    Folder:=f.ItemType=ditDirectory;
    FileSize:=f.Size;
    Readable:=Pos('x',AnsiLowerCase(f.UserPermissions))>0;
    Writable:=True;
    DateModified:=f.ModifiedDate;
    if Folder then begin
      CloseShellImageIndex:=DefaultFolderDescription.CloseImageIndex;
      OpenShellImageIndex:=DefaultFolderDescription.OpenImageIndex;
      FileTypeName:='Folder';
    end else begin
      GetFileExtensionShellDescription(ExtractFileExt(f.FileName),FileTypeName,OpenShellImageIndex);
      CloseShellImageIndex:=OpenShellImageIndex;
    end;
  end;
end;

{ TFTPFileSystem }

procedure TFTPFileSystem.Connect;
begin
  if Assigned(FLoginDialog) then begin
    FLoginDialog.Host:=Host;
    FLoginDialog.UserName:=UserName;
    FLoginDialog.PassWord:=PassWord;
    if FLoginDialog.Execute then begin
      Host:=FLoginDialog.Host;
      UserName:=FLoginDialog.UserName;
      PassWord:=FLoginDialog.PassWord;
    end else
      Exit;
  end;
  if FTransferring then
    FFTP.Abort;
  if FFTP.Connected then begin
    FFTP.Disconnect;
    NotifyHooks(hDisconnect);
  end;
  try
    FFTP.Connect;
  except
    raise;
  end;
  if FFTP.Connected then begin
    NotifyHooks(hConnect);
    CurrentDirectory:=FFTP.RetrieveCurrentDir+'/';
  end;
end;

constructor TFTPFileSystem.Create(AOwner: TComponent);
begin
  inherited;
  FCriticalSection:=TCriticalSection.Create;
  FFTP:=TidFTP.Create(Self);
  FFTP.OnDisconnected:=Disconnected;
  FFTP.OnStatus:=Status;
  FFTP.OnWork:=Work;
  FFTP.OnWorkBegin:=WorkBegin;
  FFTP.OnWorkEnd:=WorkEnd;
  FFTP.Passive:=True;
end;

procedure TFTPFileSystem.CreateDirectory(Path: string);
begin
  FFTP.MakeDir(Path);
end;

procedure TFTPFileSystem.DeleteDirectory(Path: string);
begin
  FFTP.RemoveDir(Path);
end;

procedure TFTPFileSystem.DeleteFile(FileName: string);
begin
  FFTP.Delete(FileName);
end;

destructor TFTPFileSystem.Destroy;
begin
  LoginDialog:=nil;
  Disconnect(False);
  FCriticalSection.Destroy;
  inherited;
end;

function TFTPFileSystem.DirectoryExists(Path: string): Boolean;
begin
  Result:=False;
  try
    if (Path='') or (Path='/') then
      FFTP.ChangeDir('//')
    else
      FFTP.ChangeDir(Path);
    Result:=True;
  except
  end;
end;

procedure TFTPFileSystem.Disconnect(Force: Boolean);
begin
  if Force and Connected then
    FFTP.Abort;
  if Connected then begin
    NotifyHooks(hDisconnect);
    if not (csDestroying in ComponentState) then
      FFTP.Disconnect;
  end;
  inherited;
end;

procedure TFTPFileSystem.Disconnected(Sender: TObject);
begin
  NotifyHooks(hDisconnect);
end;

procedure TFTPFileSystem.Echo(s: string);
begin
  if Assigned(FOnEcho) then
    FOnecho(Self,s);
end;

function TFTPFileSystem.FileExists(FileName: string): Boolean;
var
  Path,Name:string;
  a:Integer;
begin
  Result:=False;
  FileName:=AnsiReplaceText(FileName,'/','\');
  Path:=AnsiReplaceText(ExtractFilePath(FileName),'\','/');
  Name:=ExtractFileName(FileName);
  FileName:=AnsiReplaceText(FileName,'\','/');
  try
    if (Path='/') or (Path='') then
      FFTP.ChangeDir('//')
    else
      FFTP.ChangeDir(Path);
    FFTP.List(nil);
    with FFTP.DirectoryListing do 
      for a:=0 to Count-1 do
        Result:=Result or (Items[a].FileName=Name);
  except
  end;
end;

function TFTPFileSystem.GetConnected: Boolean;
begin
  Result:=FFTP.Connected;
end;

function TFTPFileSystem.GetDirectoryContent(Path: string): TFileItemList;
var
  a:Integer;
begin
  Result:=inherited GetDirectoryContent(Path);
  try
    FCriticalSection.Enter;
//    try
      if (Path='/') or (Path='') then
        FFTP.ChangeDir('//')
      else
        FFTP.ChangeDir(Path);
      FFTP.List(nil);
      with FFTP.DirectoryListing do begin
        for a:=0 to Count-1 do
          try
            if (Items[a].ItemType in [ditDirectory,ditFile,ditSymbolicLink]) and
               (Items[a].FileName<>'.') and
               (Items[a].FileName<>'..') and
               (Items[a].FileName<>'')  then
              Result.AddItem(FTPItemToFileItemDescriptor(Copy(Path,1,Length(Path)-1),Items[a]));
          except
            Echo('Error retrieving '+Path+' content.');
          end;
      end;
//    except
//    end;
  finally
    FCriticalSection.Leave;
  end;
end;

function TFTPFileSystem.GetHost: string;
begin
  Result:=FFTP.Host;
end;

function TFTPFileSystem.GetPassWord: string;
begin
  Result:=FFTP.Password;
end;

function TFTPFileSystem.GetUserName: string;
begin
  Result:=FFTP.Username;
end;

procedure TFTPFileSystem.Notification(AComponent: TComponent;
  Operation: TOperation);
begin
  inherited;
  if (AComponent=FLoginDialog) and (Operation=opRemove) then
    FLoginDialog:=nil;
end;

procedure TFTPFileSystem.ReadStreamFromFile(FileName: string;
  Stream: TStream);
begin
  FFTP.Get(FileName,Stream);
end;

procedure TFTPFileSystem.RenameDirectory(OldPath, NewPath: string);
begin
  FFTP.Rename(OldPath,NewPath);
end;

procedure TFTPFileSystem.RenameFile(OldFileName, NewFileName: string);
begin
  FFTP.Rename(OldFileName,NewFileName);
end;

procedure TFTPFileSystem.SetHost(const Value: string);
begin
  FFTP.Host:=Value;
end;

procedure TFTPFileSystem.SetLoginDialog(const Value: TFTPLoginDialog);
begin
  FLoginDialog:=Value;
  if Assigned(FLoginDialog) then
    FLoginDialog.FreeNotification(Self);
end;

procedure TFTPFileSystem.SetOnEcho(const Value: TEchoEvent);
begin
  FOnEcho := Value;
end;

procedure TFTPFileSystem.SetPassWord(const Value: string);
begin
  FFTP.Password:=Value;
end;

procedure TFTPFileSystem.SetUserName(const Value: string);
begin
  FFTP.Username:=Value;
end;

procedure TFTPFileSystem.Status(axSender: TObject;
  const axStatus: TIdStatus; const asStatusText: String);
begin
  Echo(asStatusText);
end;

procedure TFTPFileSystem.Work(Sender: TObject; AWorkMode: TWorkMode;
  const AWorkCount: Integer);
begin
  Echo('Work in progress...');
end;

procedure TFTPFileSystem.WorkBegin(Sender: TObject; AWorkMode: TWorkMode;
  const AWorkCountMax: Integer);
begin
  Echo('Work begin.');
end;

procedure TFTPFileSystem.WorkEnd(Sender: TObject; AWorkMode: TWorkMode);
begin
  Echo('Work done.');
end;

procedure TFTPFileSystem.WriteStreamToFile(FileName: string;
  Stream: TStream);
begin
  FFTP.Put(Stream,FileName);
end;

end.
 