unit FolderTreeExplorer;

interface

uses
  SysUtils, Classes, Controls, ComCtrls, CustomFileSystem, Dialogs;

type
  TFolderNode=class(TTreenode)
  private
    FFileItemDescriptor:TFileItemDescriptor;

    procedure SetFileItemDescriptor(Value:TFileitemDescriptor);
    function GetFolderItem(Index: Integer): TFolderNode;
    procedure SetFolderItem(Index: Integer; const Value: TFolderNode);
  public
    constructor Create(AOwner:TTreeNodes;AFileitemDescriptor:TFileItemDescriptor);

    property Descriptor:TFileItemDescriptor read FFileItemDescriptor;

    property FolderItem[Index:Integer]:TFolderNode read GetFolderItem write SetFolderItem;default;
  end;

  TFolderNodes=class(TTreeNodes)
  private
    function GetFolderNode(Index: Integer): TFolderNode;
  public
    property FolderNode[Index:Integer]:TFolderNode read GetFolderNode;default;
  end;

  TFolderTreeExplorer = class(TCustomTreeView,IFileSystemHook)
  private
    FDefaultFileSystemImplementation:TCustomFileSystemHook;
    FLockedCurrentDirectory:Boolean;

    function GetFileSystem: TCustomFileSystem;
    function GetSelected: TFolderNode;
    procedure SetSelected(const Value: TFolderNode);
    function GetFolderItems: TFolderNodes;
    procedure SetFolderItems(const Value: TFolderNodes);
  protected
    property DefaultFileSystemImplementation:TCustomFileSystemHook read FDefaultFileSystemImplementation implements IFileSystemHook;
    procedure IFileSystemHook.SetCurrentDirectory=SetCurrentDirectory;
    procedure IFileSystemHook.SetFileSystem=SetFileSystem;
    procedure IFileSystemHook.DoConnect=DoConnect;
    procedure IFileSystemHook.DoDisconnect=DoDisconnect;
    procedure IFileSystemHook.DoUpdate=DoUpdate;

    procedure SetCurrentDirectory(const Value:string);

    procedure SetFileSystem(const Value: TCustomFileSystem);

    procedure DoConnect;
    procedure DoDisconnect;
    procedure DoUpdate;

    procedure Change(Node:TTreeNode);override;
    procedure Expand(Node:TTreeNode);override;
    procedure Collapse(Node:TTreeNode);override;
  public
    constructor Create(AOwner:TComponent);override;

    function GetNodeFromPath(APath:string):TFolderNode;
    procedure UpdateFromNode(ANode:TFolderNode);

    property Selected:TFolderNode read GetSelected write SetSelected;

    property FolderItems:TFolderNodes read GetFolderItems write SetFolderItems;

    destructor Destroy;override;
  published
    property FileSystem:TCustomFileSystem read GetFileSystem write SetFileSystem;

    property Align;
    property Anchors;
    property BevelEdges;
    property BevelInner;
    property BevelOuter;
    property BevelKind default bkNone;
    property BevelWidth;
    property BiDiMode;
    property BorderStyle;
    property BorderWidth;
    property ChangeDelay;
    property Color;
    property Ctl3D;
    property Constraints;
    property DragKind;
    property DragCursor;
    property DragMode;
    property Enabled;
    property Font;
    property HideSelection;
    property HotTrack;
    property Images;
    property Indent;
    property MultiSelectStyle;
    property ParentBiDiMode;
    property ParentColor default False;
    property ParentCtl3D;
    property ParentFont;
    property ParentShowHint;
    property PopupMenu;
//    property ReadOnly;
    property RightClickSelect;
    property RowSelect;
    property ShowButtons;
    property ShowHint;
    property ShowLines;
    property ShowRoot;
    property SortType;
    property StateImages;
    property TabOrder;
    property TabStop default True;
    property ToolTips;
    property Visible;
    property OnAddition;
    property OnAdvancedCustomDraw;
    property OnAdvancedCustomDrawItem;
    property OnChange;
    property OnChanging;
    property OnClick;
    property OnCollapsed;
    property OnCollapsing;
    property OnCompare;
    property OnContextPopup;
    property OnCreateNodeClass;
    property OnCustomDraw;
    property OnCustomDrawItem;
    property OnDblClick;
    property OnDeletion;
    property OnDragDrop;
    property OnDragOver;
    property OnEdited;
    property OnEditing;
    property OnEndDock;
    property OnEndDrag;
    property OnEnter;
    property OnExit;
    property OnExpanding;
    property OnExpanded;
    property OnGetImageIndex;
    property OnGetSelectedIndex;
    property OnKeyDown;
    property OnKeyPress;
    property OnKeyUp;
    property OnMouseDown;
    property OnMouseMove;
    property OnMouseUp;
    property OnStartDock;
    property OnStartDrag;
  end;

implementation

{ TFolderTreeExplorer }

procedure TFolderTreeExplorer.Change(Node: TTreeNode);
begin
  inherited;
  if Assigned(Selected) and not FLockedCurrentDirectory then
    with Selected do
      SetCurrentDirectory(FFileItemDescriptor.FilePath+FFileItemDescriptor.FileName+'/');
  UpdateFromNode(TFolderNode(Node));
end;

procedure TFolderTreeExplorer.Collapse(Node: TTreeNode);
begin
  inherited Collapse(Node);
  Node.ImageIndex:=TFolderNode(Node).FFileItemDescriptor.CloseShellImageIndex;
end;

constructor TFolderTreeExplorer.Create(AOwner: TCOmponent);
begin
  inherited;
  ReadOnly:=True;
  AutoExpand:=True;
  HideSelection:=False;
  FDefaultFileSystemImplementation:=TCustomFileSystemHook.Create(Self);
  Images:=ShellSmallImageList;
  Items.Destroy;
  Items:=TFolderNodes.Create(Self);
end;

destructor TFolderTreeExplorer.Destroy;
begin
  inherited;
end;

procedure TFolderTreeExplorer.DoConnect;
begin
  Items.Clear;
end;

procedure TFolderTreeExplorer.DoDisconnect;
begin
  Items.Clear;
end;

procedure TFolderTreeExplorer.DoUpdate;
begin
  FLockedCurrentDirectory:=True;
  if Items.Count>0 then
    UpdateFromNode(FolderItems[0]);
  FLockedCurrentDirectory:=False;
end;

procedure TFolderTreeExplorer.Expand(Node: TTreeNode);
begin
  inherited;
  Node.ImageIndex:=TFolderNode(Node).FFileItemDescriptor.OpenShellImageIndex;
  UpdateFromNode(TFolderNode(Node));
end;

function TFolderTreeExplorer.GetFileSystem: TCustomFileSystem;
begin
  Result:=FDefaultFileSystemImplementation.FileSystem;
end;

function TFolderTreeExplorer.GetFolderItems: TFolderNodes;
begin
  Result:=TFolderNodes(inherited Items);
end;

function TFolderTreeExplorer.GetNodeFromPath(APath: string): TFolderNode;
var
  a:Integer;
begin
  Result:=nil;
  if Items.Count>0 then begin
    for a:=0 to Items.Count-1 do
      with FolderItems[a].FFileItemDescriptor do
        if Pos(FilePath+FileName+'/',APath)=1 then begin
          Result:=FolderItems[a];
          break;
        end;
    while Assigned(Result) and (APath<>Result.FFileItemDescriptor.FilePath+Result.FFileItemDescriptor.FileName+'/') do
      for a:=Result.Count-1 downto -1 do
        if a<0 then
          Result:=nil
        else begin
          with Result.FolderItem[a].FFileItemDescriptor do
            if Pos(FilePath+FileName+'/',APath)=1 then begin
              Result:=Result.FolderItem[a];
              Break;
            end;
        end;
  end;
end;

function TFolderTreeExplorer.GetSelected: TFolderNode;
begin
  Result:=TFolderNode(inherited Selected);
end;

procedure TFolderTreeExplorer.SetCurrentDirectory(const Value: string);
var
  s,t,s1,s2:string;
  FirstNode,LastNode,n:TFolderNode;
  d:TFileItemDescriptor;
begin
  if FLockedCurrentDirectory then
    Exit;
  FLockedCurrentDirectory:=True;
  if Assigned(FDefaultFileSystemImplementation.FileSystem) then
    FileSystem.CurrentDirectory:=Value;
  s:=Value;
  t:='';
  LastNode:=nil;
  Firstnode:=nil;
  repeat
    SplitString(s,s1,s2,['/'],False);
    n:=GetNodeFromPath(t+s1+'/');
    if not Assigned(n) then begin
      d.FileName:=s1;
      if t='' then begin
        d.FileName:='';
        d.FilePath:='';
//        showmessage(s1);
//        d.FilePath:=s1;
//        d.FileName:='';
      end else
        d.FilePath:=t;
      d.Folder:=True;
      d.OpenShellImageIndex:=DefaultFolderDescription.OpenImageIndex;
      d.CloseShellImageIndex:=DefaultFolderDescription.CloseImageIndex;
      n:=TFolderNode.Create(Items,d);
      LastNode:=TFolderNode(Items.AddNode(n,LastNode,n.Text,nil,naAddChild));
      n.SetFileItemDescriptor(d);
      if not Assigned(FirstNode) then
        FirstNode:=LastNode;
    end else
      LastNode:=n;
    t:=t+s1+'/';
    s:=s2;
  until s2='';
  UpdateFromNode(FirstNode);
  LastNode:=GetNodeFromPath(Value);
  if Assigned(LastNode) and not LastNode.Selected then begin
    Expand(LastNode);
//    LastNode.Expand(False);
    Select([Lastnode]);
  end;
  FLockedCurrentDirectory:=False;
end;

procedure TFolderTreeExplorer.SetFileSystem(
  const Value: TCustomFileSystem);
begin
  FDefaultFileSystemImplementation.FileSystem:=Value;
end;

procedure TFolderTreeExplorer.SetFolderItems(const Value: TFolderNodes);
begin
  Items:=Value;
end;

procedure TFolderTreeExplorer.SetSelected(const Value: TFolderNode);
begin
  inherited Selected:=Value;
end;

procedure TFolderTreeExplorer.UpdateFromNode(ANode: TFolderNode);
var
  a,b:Integer;
  n:TFolderNode;
  t:Boolean;
begin
  if (not Assigned(ANode)) or //(not ANode.FFileItemDescriptor.Readable) or
     ((ANode.Count>0) and not ANode.Expanded) then
    Exit;
  with ANode.FFileItemDescriptor do
    with FileSystem.GetDirectoryContent(FilePath+FileName+'/') do begin
      for a:=0 to Count-1 do begin
        if Items[a].Folder then begin
          t:=False;
          for b:=0 to ANode.Count-1 do begin
            t:=t or (ANode.FolderItem[b].FFileItemDescriptor.FileName=Items[a].FileName);
            if t then begin
              ANode.FolderItem[b].SetFileItemDescriptor(Items[a]);
              Break;
            end;
          end;
          if not t then begin
            n:=TFolderNode.Create(Self.Items,Items[a]);
            Self.Items.AddNode(n,ANode,n.Text,nil,naAddChild);
            n.SetFileItemDescriptor(Items[a]);
            n.Expand(False);
          end;
        end;
      end;
      for a:=ANode.Count-1 downto 0 do begin
        t:=False;
        for b:=Count-1 downto 0 do begin
          t:=t or (Items[b].FileName=ANode.FolderItem[a].FFileItemDescriptor.FileName);
          if t then begin
            DeleteItem(b);
            Break;
          end;
        end;
        if not t then
          ANode.Item[a].Destroy
        else begin
          with ANode.FolderItem[a] do
            if {FFileItemDescriptor.Readable and }(Selected or Expanded) or
               ((Count=0) and ANode.FolderItem[a].FFileItemDescriptor.FastExplore and (ANode.Selected or ANode.Expanded)) then
              UpdateFromNode(ANode.FolderItem[a]);
        end;
      end;
      Destroy;
    end;
end;

{ TFolderNode }

constructor TFolderNode.Create(AOwner: TTreeNodes;
  AFileitemDescriptor: TFileItemDescriptor);
begin
  inherited Create(AOwner);
//  SetFileitemDescriptor(AFileitemDescriptor);
end;

function TFolderNode.GetFolderItem(Index: Integer): TFolderNode;
begin
  Result:=TFolderNode(Item[Index]);
end;

procedure TFolderNode.SetFileItemDescriptor(Value: TFileitemDescriptor);
begin
  FFileItemDescriptor:=Value;
  if FFileItemDescriptor.FileName='' then begin
    FFileItemDescriptor:=TFolderTreeExplorer(Owner.Owner).FileSystem.RootRecordDescription;
    Text:=FFileItemDescriptor.FileTypeName;
  end else
    Text:=FFileItemDescriptor.FileName;
  ImageIndex:=FFileItemDescriptor.CloseShellImageIndex;
  SelectedIndex:=FFileItemDescriptor.OpenShellImageIndex;
//  Text:=Text+'         '+FFileItemDescriptor.FilePath+'|'+FFileItemDescriptor.FileName;
end;

procedure TFolderNode.SetFolderItem(Index: Integer;
  const Value: TFolderNode);
begin
  Item[Index]:=Value;
end;

{ TFolderNodes }

function TFolderNodes.GetFolderNode(Index: Integer): TFolderNode;
begin
  Result:=TFolderNode(Item[Index]);
end;

end.
