unit PipeConnection;

interface

uses
  SysUtils,Windows,UnitHandler,Classes;

type
  TPipeConnectionServer=class(TCustomConnectionServer)
  private
    FListenThread:TThread;
    FPipe:THandle;
    FEvent:THandle;

    procedure ExecuteListen;
  protected
    function CheckValidAddress(const Value:string):Boolean;override;
    function GetConnected:Boolean;override;
  public
    constructor Create(AOwner:TComponent);override;

    procedure Connect;override;
    procedure Disconnect;override;

    destructor Destroy;override;
  published
  end;

  TPipeConnectionClient=class(TCustomConnectionClient)
  private
    FPipe:THandle;
  protected
    function CheckValidAddress(const Value:string):Boolean;override;
  public
    procedure Connect;override;
  published
  end;

  TPipeConnection=class(TConnection)
  private
    FPipe:THandle;
  protected
    procedure WriteCardinal(const Value:Cardinal);override;
    function ReadCardinal:Cardinal;override;

    procedure SubWriteBuffer(p:TStream;var Count:Cardinal);override;
    procedure SubReadBuffer(p:TStream;var Count:Cardinal);override;

    procedure Disconnect;override;
  public
    constructor Create(AConnectionComponent:TConnectionComponent;AsServer:Boolean;APipe:THandle);

    destructor Destroy;override;
  end;

implementation

function CheckPipeAddress(Value:string):Boolean;
begin
  Result:=True;
end;

{ TPipeConnectionServer }

function TPipeConnectionServer.CheckValidAddress(
  const Value: string): Boolean;
begin
  Result:=CheckPipeAddress(Value);
end;

procedure TPipeConnectionServer.Connect;
begin
  inherited;
  FListenThread:=TProcThread.Create(ExecuteListen,False);
end;

constructor TPipeConnectionServer.Create(AOwner: TComponent);
begin
  inherited;
  FEvent:=CreateEvent(nil,False,False,nil);
end;

destructor TPipeConnectionServer.Destroy;
begin

  inherited;
end;

procedure TPipeConnectionServer.Disconnect;
var
  h:THandle;
begin
  inherited;
  if Assigned(FListenThread) then begin
    FTerminated:=True;
    h:=CreateFile(PChar(FAddress),GENERIC_READ or GENERIC_WRITE,0,nil,OPEN_EXISTING,0,0);
    CloseHandle(h);
    FListenThread.WaitFor;
    FListenThread.Destroy;
    FListenThread:=nil;
  end;
end;

procedure TPipeConnectionServer.ExecuteListen;
begin
  repeat
    FPipe:=CreateNamedPipe(PChar(FAddress),
                          PIPE_ACCESS_DUPLEX,
                          PIPE_TYPE_MESSAGE or PIPE_READMODE_MESSAGE or PIPE_WAIT,
                          PIPE_UNLIMITED_INSTANCES,
                          0,0,1000,nil);
    Assert(FPipe<>0,SysErrorMessage(GetLastError));
    if ConnectNamedPipe(FPipe,nil) then begin
      if not FTerminated then
        DoConnect(TPipeConnection.Create(Self,True,FPipe))
      else
        CloseHandle(FPipe);
    end else
      CloseHandle(FPipe);
  until FTerminated;
end;

function TPipeConnectionServer.GetConnected: Boolean;
begin
  Result:=Assigned(FListenThread);
end;

{ TPipeConnectionClient }

function TPipeConnectionClient.CheckValidAddress(
  const Value: string): Boolean;
begin
  Result:=CheckPipeAddress(Value);
end;

procedure TPipeConnectionClient.Connect;
var
  t:Boolean;
  n:Cardinal;
begin
  t:=False;
  repeat
    FPipe:=CreateFile(PChar(FAddress),
                     GENERIC_READ or GENERIC_WRITE,
                     0,nil,OPEN_EXISTING,0,0);
    if FPipe=INVALID_HANDLE_VALUE then begin
      n:=GetLastError;
      if n=ERROR_PIPE_BUSY then
        WaitNamedPipe(PChar(Name),NMPWAIT_USE_DEFAULT_WAIT)
      else
        Break;
    end else
      t:=True;
  until t;
  if not t then
    raise Exception.Create(SysErrorMessage(GetLastError));
  DoConnect(TPipeConnection.Create(Self,False,FPipe));
end;

{ TPipeConnection }

constructor TPipeConnection.Create(
  AConnectionComponent: TConnectionComponent; AsServer: Boolean;
  APipe: THandle);
begin
  FPipe:=APipe;
  inherited Create(AConnectionComponent,AsServer,nil);
end;

destructor TPipeConnection.Destroy;
begin
  inherited;
  if FPipe<>0 then begin
    DisconnectNamedPipe(FPipe);
    CloseHandle(FPipe);
  end;
end;

procedure TPipeConnection.Disconnect;
begin
  _AddRef;
  DisconnectNamedPipe(FPipe);
  CloseHandle(FPipe);
  FPipe:=0;
  inherited Disconnect;
  _Release;
end;

function TPipeConnection.ReadCardinal: Cardinal;
var
  n:Cardinal;
begin
  if not ReadFile(FPipe,Result,SizeOf(Result),n,nil) then
    raise Exception.Create(SysErrorMessage(GetLastError));
  if n<>SizeOf(Result) then
    Assert(False,FConnectionComponent.ClassName);
end;

procedure TPipeConnection.SubReadBuffer(p: TStream; var Count: Cardinal);
var
  Data:array[1..1024] of Byte;
  n,s,r:Cardinal;
begin
  n:=ReadCardinal;
  if n<Count then
    Count:=0;
  WriteCardinal(Count);
  if Count>0 then
    repeat
      Assert(PeekNamedPipe(FPipe,@Data[1],1024,@n,@s,@r),SysErrorMessage(GetLastError));
      ReadFile(FPipe,Data[1],n,n,nil);
      p.Write(Data[1],n);
    until s=Count;
end;

procedure TPipeConnection.SubWriteBuffer(p: TStream; var Count: Cardinal);
var
  Data:Pointer;
  n:Cardinal;
begin
  GetMem(Data,Count);
  p.Read(Data^,Count);
  WriteCardinal(Count);
  n:=ReadCardinal;
  Count:=n;
  if Count>0 then
    Assert(WriteFile(FPipe,Data^,Count,n,nil),SysErrorMessage(GetLastError));
  Assert(n=Count);
  FreeMem(Data);
end;

procedure TPipeConnection.WriteCardinal(const Value: Cardinal);
var
  n:Cardinal;
begin
  if not WriteFile(FPipe,Value,SizeOf(Value),n,nil) then
    raise Exception.Create(SysErrorMessage(GetLastError));
  if n<>SizeOf(Value) then
    Assert(False);
end;

end.
