unit SmartEdit;

interface

uses
  Windows, Messages, SysUtils, Classes, Controls, StdCtrls, Dialogs;

type
  TSmartRecord=record
    LastUsage:TDateTime;
    UsageCount:Integer;
  end;

  PSmartRecord=^TSmartRecord;

  TSmartStrategy=(ssLastUsed,ssMoreUsed);

  TSmartEdit = class(TCustomComboBox)
  private
    FPersistent:Boolean;
    FStorageFile:string;
    FStoredStrings:TStringList;
    FStrategy:TSmartStrategy;

    procedure SetPersistent(const Value: Boolean);
    procedure SetStorageFile(const Value: string);
  protected
    procedure DoExit;override;
    procedure KeyUp(var Key:Word;Shift:TShiftState);override;

    procedure AdjustDropDown;override;

    procedure UpdateList;
  public
    constructor Create(AOwner:TComponent);override;

    function GetBestMatch(s:string):string;

    procedure LoadStorage;
    procedure SaveStorage;
    procedure AddToStorage(Entry:string;LastUsage:TDateTime=0;UsageCount:Integer=1);
    procedure RemoveFromStorage(Entry:string);
    procedure ResetStorage;
    procedure Store;

    destructor Destroy;override;
  published
    property Persistent:Boolean read FPersistent write SetPersistent;
    property StorageFile:string read FStorageFile write SetStorageFile;
    property Strategy:TSmartStrategy read FStrategy write FStrategy;
    property Text;

    property AutoCloseUp default False;
    property BevelEdges;
    property BevelInner;
    property BevelKind default bkNone;
    property BevelOuter;
    property Anchors;
    property BiDiMode;
    property CharCase;
    property Color;
    property Constraints;
    property Ctl3D;
    property DragCursor;
    property DragKind;
    property DragMode;
    property DropDownCount;
    property Enabled;
    property Font;
    property ImeMode;
    property ImeName;
    property ItemHeight;
    property MaxLength;
    property ParentBiDiMode;
    property ParentColor;
    property ParentCtl3D;
    property ParentFont;
    property ParentShowHint;
    property PopupMenu;
    property ShowHint;
    property Sorted;
    property TabOrder;
    property TabStop;
    property Visible;
    property OnChange;
    property OnClick;
    property OnCloseUp;
    property OnContextPopup;
    property OnDblClick;
    property OnDragDrop;
    property OnDragOver;
    property OnDrawItem;
    property OnDropDown;
    property OnEndDock;
    property OnEndDrag;
    property OnEnter;
    property OnExit;
    property OnKeyDown;
    property OnKeyPress;
    property OnKeyUp;
    property OnMeasureItem;
    property OnSelect;
    property OnStartDock;
    property OnStartDrag;
  end;

implementation

{ TSmartEdit }

procedure TSmartEdit.AddToStorage(Entry: string; LastUsage: TDateTime; UsageCount: Integer);
var
  a:Integer;
  p:PSmartRecord;
begin
  if Entry='' then
    Exit;
  a:=FStoredStrings.IndexOf(Entry);
  if a=-1 then begin
    New(p);
    p^.UsageCount:=0;
    a:=FStoredStrings.AddObject(Entry,TObject(p));
  end;
  Inc(PSmartRecord(FStoredStrings.Objects[a])^.UsageCount,UsageCount);
  if LastUsage=0 then
    PSmartRecord(FStoredStrings.Objects[a])^.LastUsage:=Now
  else
    PSmartRecord(FStoredStrings.Objects[a])^.LastUsage:=LastUsage;
end;

procedure TSmartEdit.AdjustDropDown;
begin
  UpdateList;
  inherited;
end;

constructor TSmartEdit.Create(AOwner: TComponent);
begin
  inherited;
  AutoComplete:=False;
  FStoredStrings:=TStringList.Create;
  FStorageFile:='UserData.txt';
  Sorted:=True;
end;

destructor TSmartEdit.Destroy;
begin
  Store;
  if FPersistent then
    SaveStorage;
  ResetStorage;
  FStoredStrings.Destroy;
  inherited;
end;

procedure TSmartEdit.doExit;
begin
  inherited;
  Store;
end;

function TSmartEdit.GetBestMatch(s: string): string;
var
  a:Integer;
begin
  Result:='';
  for a:=0 to FStoredStrings.Count-1 do
    if s=Copy(FStoredStrings[a],1,Length(s)) then
      Result:=FStoredStrings[a];
end;

procedure TSmartEdit.KeyUp(var Key: Word; Shift: TShiftState);
var
  s,t:string;
begin
  if IsCharAlphaNumeric(Char(Key)) then begin
    s:=Copy(Text,1,SelStart);
    t:=GetBestMatch(s);
    if (t<>'') and (s<>'') then begin
      DroppedDown:=True;
      Text:=t;
      SelStart:=Length(s);
      SelLength:=Length(t)-Length(s);
      Key:=0;
    end;
  end else;
    UpdateList;
  if Text='' then
    DroppedDown:=False;
  inherited;
end;

procedure TSmartEdit.LoadStorage;
var
  l:TStringList;
  a,b,c:Integer;
begin
  if FileExists(FStorageFile) then begin
    l:=TStringList.Create;
    l.LoadFromFile(FStorageFile);
    a:=l.IndexOf('['+Name+']');
    if a<>-1 then begin
      b:=StrToInt(l[a+1]);
      for c:=0 to b-1 do
        AddToStorage(l[a+2+3*c],StrToDateTime(l[a+3+3*c]),StrToInt(l[a+4+3*c]));
    end;
    l.Destroy;
  end;
end;

procedure TSmartEdit.RemoveFromStorage(Entry: string);
var
  a:Integer;
begin
  a:=FStoredStrings.IndexOf(Entry);
  if a>-1 then begin
    Dispose(PSmartRecord(FStoredStrings.Objects[a]));
    FStoredStrings.Delete(a);
  end;
end;

procedure TSmartEdit.ResetStorage;
begin
  while FStoredStrings.Count>0 do
    RemoveFromStorage(FStoredStrings[0]);
end;

procedure TSmartEdit.SaveStorage;
var
  l:TStringList;
  a,b,c:Integer;
begin
  if (csDesigning in ComponentState) then
    Exit;
  l:=TStringList.Create;
  if FileExists(FStorageFile) then begin
    l.LoadFromFile(FStorageFile);
    a:=l.IndexOf('['+Name+']');
    if a<>-1 then begin
      b:=StrToInt(l[a+1]);
      for c:=0 to 3*b do
        l.Delete(a+1);
    end else
      a:=l.Add('['+Name+']');
  end else
    a:=l.Add('['+Name+']');
  l.Insert(a+1,IntToStr(FStoredStrings.Count));
  for b:=0 to FStoredStrings.Count-1 do begin
    l.Insert(a+2+3*b,FStoredStrings[b]);
    l.Insert(a+3+3*b,DateTimeToStr(PSmartRecord(FStoredStrings.Objects[b])^.LastUsage));
    l.Insert(a+4+3*b,IntToStr(PSmartRecord(FStoredStrings.Objects[b])^.UsageCount));
  end;
  l.SaveToFile(FStorageFile);
end;

procedure TSmartEdit.SetPersistent(const Value: Boolean);
begin
  if FPersistent xor Value then begin
    FPersistent:=Value;
    if FPersistent then
      LoadStorage;
  end;
end;

procedure TSmartEdit.SetStorageFile(const Value: string);
begin
  if FStorageFile<>Value then begin
    FStorageFile:=Value;
    LoadStorage;
  end;
end;

procedure TSmartEdit.Store;
begin
  AddToStorage(Text);
end;

procedure TSmartEdit.UpdateList;
var
  a,b,c:Integer;
  s,t:string;
begin
  t:=Text;
  b:=SelStart;
  c:=SelLength;
  if SelLength=0 then
    s:=Text
  else
    s:=Copy(Text,1,SelStart);
  Items.Clear;
  for a:=0 to FStoredStrings.Count-1 do
    if s=Copy(FStoredStrings[a],1,Length(s)) then
      Items.Add(FStoredStrings[a]);
  Text:=t;
  SelStart:=b;
  SelLength:=c;
end;

end.
