unit TCPConnection;

interface

uses
  Classes, SysUtils, IdBaseComponent, IdComponent, IdTCPConnection, IdTCPClient,
  IdThreadMgr, IdThreadMgrDefault, IdTCPServer, UnitHandler, Windows;

type
  TTCPConnectionServer=class(TCustomConnectionServer)
  private
    FIdServer:TIdTCPServer;

    procedure TCPServerConnect(AThread:TIdPeerThread);
    procedure TCPServerExecute(AThread:TIdPeerThread);
    procedure TCPServerDisconnect(AThread:TIdPeerThread);
    procedure TCPServerException(AThread:TIdPeerThread;AException:Exception);
  protected
    function CheckValidAddress(const Value:string):Boolean;override;
    function GetConnected:Boolean;override;
  public
    constructor Create(AOwner:TComponent);override;

    procedure Connect;override;
    procedure Disconnect;override;

    destructor Destroy;override;
  published
  end;

  TTCPConnectionClient=class(TCustomConnectionClient)
  private
    FIdClient:TIdTCPClient;

    procedure TCPClientConnect(Sender:TObject);
    procedure TCPClientDisconnect(Sender:TObject);
  protected
    function CheckValidAddress(const Value:string):Boolean;override;
  public
    constructor Create(AOwner:TComponent);override;

    procedure Connect;override;
  published
  end;

  TTCPConnection=class(TConnection)
  private
    FConnection:TIdTCPConnection;
  protected
    procedure WriteCardinal(const Value:Cardinal);override;
    function ReadCardinal:Cardinal;override;

    procedure SubWriteBuffer(p:TStream;var Count:Cardinal);override;
    procedure SubReadBuffer(p:TStream;var Count:Cardinal);override;

    procedure Disconnect;override;

    function GetRemoteAddress:string;override;
  public
    constructor Create(AConnectionComponent:TConnectionComponent;AsServer:Boolean;AThread:TIdPeerThread);

    destructor Destroy;override;
  end;

implementation

uses IdSocketHandle;

{ TTCPConnectionServer }

function TTCPConnectionServer.CheckValidAddress(
  const Value: string): Boolean;
begin
  Result:=True;
end;

procedure TTCPConnectionServer.Connect;
begin
  if FIdServer.Active then
    Exit;
  FIdServer.Bindings.Clear;
  with FIdServer.Bindings.Add do begin
    IP:=FAddress;
    Port:=6767;
  end;
  FIdServer.Active:=True;
end;

constructor TTCPConnectionServer.Create(AOwner: TComponent);
begin
  inherited;
  FIdServer:=TIdTCPServer.Create(Self);
  FIdServer.OnConnect:=TCPServerConnect;
  FIdServer.OnExecute:=TCPServerExecute;
  FIdServer.OnDisconnect:=TCPServerDisconnect;
  FIdServer.OnException:=TCPServerException;
end;

destructor TTCPConnectionServer.Destroy;
begin
  inherited Destroy;
end;

procedure TTCPConnectionServer.Disconnect;
begin
//  inherited Disconnect;
  FIdServer.Active:=False;
end;

function TTCPConnectionServer.GetConnected: Boolean;
begin
  Result:=FIdServer.Active;
end;

procedure TTCPConnectionServer.TCPServerConnect(AThread: TIdPeerThread);
begin
  AThread.Data:=TTCPConnection.Create(Self,True,AThread);
  TTCPConnection(AThread.Data)._AddRef;
  DoConnect(TTCPConnection(AThread.Data));
end;

procedure TTCPConnectionServer.TCPServerDisconnect(AThread: TIdPeerThread);
var
  c:TTCPConnection;
begin
  LockedConnectionList;
  c:=TTCPConnection(AThread.Data);
  Assert(c.FThread.ThreadId=GetCurrentThreadId);
  AThread.Data:=nil;
  try
    DoDisconnect(c);
  finally
    UnLockConnectionList;
    c._Release;
  end;
end;

procedure TTCPConnectionServer.TCPServerException(AThread: TIdPeerThread;
  AException: Exception);
begin
  Beep(440,1000);
end;

procedure TTCPConnectionServer.TCPServerExecute(AThread: TIdPeerThread);
begin
  try
    if not TTCPConnection(AThread.Data).FTerminated then
      TTCPConnection(AThread.Data).SubExecute;
  except
    if AThread.Connection.Connected then
      AThread.Connection.Disconnect;
  end;
end;

{ TTCPConnectionClient }

function TTCPConnectionClient.CheckValidAddress(
  const Value: string): Boolean;
begin
  Result:=True;
end;

procedure TTCPConnectionClient.Connect;
begin
  inherited;
  FIdClient.Host:=FAddress;
  FIdClient.Port:=6767;
  FIdClient.Connect;
end;

constructor TTCPConnectionClient.Create(AOwner: TComponent);
begin
  inherited;
  FIdClient:=TIdTCPClient.Create(Self);
  FIdClient.OnConnected:=TCPClientConnect;
  FIdClient.OnDisconnected:=TCPClientDisconnect;
end;

procedure TTCPConnectionClient.TCPClientConnect(Sender: TObject);
begin
  DoConnect(TTCPConnection.Create(Self,False,nil));
end;

procedure TTCPConnectionClient.TCPClientDisconnect(Sender: TObject);
begin

end;

{ TTCPConnection }

constructor TTCPConnection.Create(
  AConnectionComponent: TConnectionComponent; AsServer: Boolean;
  AThread: TIdPeerThread);
begin
  FPData:=TMemoryStream.Create;
  FQData:=TMemoryStream.Create;
  if Assigned(AThread) then
    FConnection:=AThread.Connection
  else
    FConnection:=TTCPConnectionClient(AConnectionComponent).FIdClient;
  inherited Create(AConnectionComponent,AsServer,AThread);
  FProgressThread:=TProcThread.Create(ExecuteProgress,False);
end;

destructor TTCPConnection.Destroy;
begin
  FTerminated:=True;
  FProgressThread.WaitFor;
  inherited Destroy;
end;

procedure TTCPConnection.Disconnect;
begin
  FConnection.Disconnect;
  if FThread.ThreadID<>GetCurrentThreadId then
    FThread.WaitFor;
end;

function TTCPConnection.GetRemoteAddress: string;
begin
  Result:=FConnection.LocalName;
end;

function TTCPConnection.ReadCardinal: Cardinal;
begin
  Result:=FConnection.ReadCardinal;
end;

procedure TTCPConnection.SubReadBuffer(p: TStream; var Count: Cardinal);
var
  q:Pointer;
begin
  if ReadCardinal<Count then
    Count:=0;
  WriteCardinal(Count);
  GetMem(q,Count);
  FConnection.ReadBuffer(q^,Count);
  p.WriteBuffer(q^,Count);
  FreeMem(q);
  FConnection.WriteLn;
end;

procedure TTCPConnection.SubWriteBuffer(p: TStream; var Count: Cardinal);
var
  q:Pointer;
begin
  WriteCardinal(Count);
  Count:=ReadCardinal;
  if Count=0 then
    p.Size:=0;
  FConnection.OpenWriteBuffer;
  GetMem(q,Count);
  p.ReadBuffer(q^,Count);
  FConnection.WriteBuffer(q^,Count);
  FreeMem(q);
  FConnection.CloseWriteBuffer;
  FConnection.ReadLn;
end;

procedure TTCPConnection.WriteCardinal(const Value: Cardinal);
begin
  FConnection.WriteCardinal(Value);
end;

end.
