unit VirtualFileSystem;

interface

uses
  SysUtils, Classes, CustomFileSystem, Contnrs, Dialogs;

type
  TDataType=(dtString,dtByte,dtWord,dtCardinal,dtInt64);

  TVirtualFileSystem = class;
  TVirtualFileItem=class;
  TVirtualFolderItem=class;

  TVirtualItem=class
  protected
    FDescriptor:TFileItemDescriptor;
    FParent:TVirtualFolderItem;
  public
    constructor Create(AParent:TVirtualFolderItem;APath:string;AName:string);

    procedure SetName(AName:string);

    function FindFile(Name:string):TVirtualFileItem;virtual;abstract;

    procedure ReadFromStream(s:TStream);virtual;
    procedure WriteToStream(s:TStream);virtual;

    destructor Destroy;override;
  end;

  TVirtualFileItem=class(TVirtualItem)
  protected
    FData:Pointer;
    FDataLength:Integer;
  public
    constructor Create(AParent:TVirtualFolderItem;APath:string;AName:string;AData:TStream);

    procedure SetData(AData:TStream);

    function FindFile(Name:string):TVirtualFileItem;override;

    procedure ReadFromStream(s:TStream);override;
    procedure WriteToStream(s:TStream);override;

    destructor Destroy;override;
  end;

  TVirtualFolderItem=class(TVirtualItem)
  protected
    FSubItems:TObjectList;
  public
    constructor Create(AParent:TVirtualFolderItem;APath:string;AName:string);

    function FindFolder(Path:string):TVirtualFolderItem;
    function FindFile(Name:string):TVirtualFileItem;override;

    procedure ReadFromStream(s:TStream);override;
    procedure WriteToStream(s:TStream);override;

    destructor Destroy;override;
  end;

  TVirtualFileSystem = class(TCustomFileSystem)
  private
    FRoot:TVirtualFolderItem;
  protected
  public
    constructor Create(AOwner:TComponent);override;

    procedure Connect;override;
    procedure Disconnect(Force:Boolean=False);override;
    function GetConnected:Boolean;override;

    function GetDirectoryContent(Path:string):TFileItemList;override;
    function DirectoryExists(Path:string):Boolean;override;
    procedure CreateDirectory(Path:string);override;
    procedure DeleteDirectory(Path:string);override;
    procedure RenameDirectory(OldPath,NewPath:string);override;

    function FileExists(FileName:string):Boolean;override;
    procedure RenameFile(OldFileName,NewFileName:string);override;
    procedure DeleteFile(FileName:string);override;
    procedure WriteStreamToFile(FileName:string;Stream:TStream);override;
    procedure ReadStreamFromFile(FileName:string;Stream:TStream);override;

    procedure WriteFileSystemToStream(s:TStream);
    procedure ReadFileSystemFromStream(s:TStream);
    procedure ClearFileSystem;

    procedure LoadVars(Vars:array of Pointer;Types:array of TDataType;Paths:array of string);
    procedure SaveVars(Vars:array of Pointer;Types:array of TDataType;Paths:array of string);

    destructor Destroy;override;
  published
  end;

implementation

{ TVirtualFileSystem }

procedure TVirtualFileSystem.ClearFileSystem;
begin
  FRoot.Destroy;
  FRoot:=TVirtualFolderItem.Create(nil,'','');
  CurrentDirectory:='/';
  NotifyHooks(hUpdate);
end;

procedure TVirtualFileSystem.Connect;
begin
  inherited;
  NotifyHooks(hConnect);
  CurrentDirectory:='/';
end;

constructor TVirtualFileSystem.Create(AOwner: TComponent);
begin
  inherited;
  FRoot:=TVirtualFolderItem.Create(nil,'','');
end;

procedure TVirtualFileSystem.CreateDirectory(Path: string);
var
  s1,s2:string;
  t:TVirtualFolderItem;
begin
  inherited;
  Assert((Path<>'') and (Path[1]='/'));
  if Path[Length(Path)]='/' then
    Path:=Copy(Path,1,Length(Path)-1);
  if not DirectoryExists(Path+'/') then begin
    SplitString(Path,s1,s2,['/'],True);
    t:=FRoot.FindFolder(s1+'/');
    if not Assigned(t) then begin
      CreateDirectory(s1+'/');
      t:=FRoot.FindFolder(s1+'/');
    end;
    Assert(Assigned(t));
    TVirtualFolderItem.Create(t,s1+'/',s2);
  end;
end;

procedure TVirtualFileSystem.DeleteDirectory(Path: string);
var
  t:TVirtualFolderItem;
begin
  inherited;
  t:=FRoot.FindFolder(Path);
  Assert(Assigned(t));
  t.Destroy;
end;

procedure TVirtualFileSystem.DeleteFile(FileName: string);
var
  t:TVirtualFileItem;
begin
  inherited;
  t:=FRoot.FindFile(FileName);
  Assert(Assigned(t));
  t.Destroy;
end;

destructor TVirtualFileSystem.Destroy;
begin
  Disconnect;
  FRoot.Destroy;
  inherited;
end;

function TVirtualFileSystem.DirectoryExists(Path: string): Boolean;
begin
  Result:=(Path='/') or (Path='//') or Assigned(FRoot.FindFolder(Path));
end;

procedure TVirtualFileSystem.Disconnect(Force: Boolean);
begin
  NotifyHooks(hDisconnect);
  inherited;
end;

function TVirtualFileSystem.FileExists(FileName: string): Boolean;
begin
  Result:=Assigned(FRoot.FindFile(FileName));
end;

function TVirtualFileSystem.GetConnected: Boolean;
begin
  Result:=FCurrentDirectory<>'';
end;

function TVirtualFileSystem.GetDirectoryContent(
  Path: string): TFileItemList;
var
  a:Integer;
  t:TVirtualFolderItem;
begin
  Result:=inherited GetDirectoryContent(Path);
  t:=FRoot.FindFolder(Path);
  if Assigned(t) then
    for a:=0 to t.FSubItems.Count-1 do
      Result.AddItem(TVirtualItem(t.FSubItems[a]).FDescriptor);
end;

function GetDataSize(t:TDataType):Integer;
begin
  Result:=0;
  case t of
    dtByte:Result:=SizeOf(Byte);
    dtWord:Result:=SizeOf(Word);
    dtCardinal:Result:=SizeOf(Cardinal);
    dtInt64:Result:=SizeOf(Int64);
  end;
end;

procedure TVirtualFileSystem.LoadVars(Vars: array of Pointer;
  Types: array of TDataType; Paths: array of string);
var
  a:Integer;
  s:TStream;
begin
  for a:=Low(Vars) to High(Vars) do begin
    case Types[a] of
      dtString:s:=TStringStream.Create('');
    else
      s:=TMemoryStream.Create;
    end;
    ReadStreamFromFile(Paths[a],s);
    s.Position:=0;
    case Types[a] of
      dtString:PString(Vars[a])^:=TStringStream(s).DataString;
    else
      s.Read(Vars[a]^,GetDataSize(Types[a]));
    end;
    s.Destroy;
  end;
end;

procedure TVirtualFileSystem.ReadFileSystemFromStream(s: TStream);
begin
  FRoot.Destroy;
  FRoot:=TVirtualFolderItem.Create(nil,'','');
  FRoot.ReadFromStream(s);
  UpdateHooks;
end;

procedure TVirtualFileSystem.ReadStreamFromFile(FileName: string;
  Stream: TStream);
var
  t:TVirtualFileItem;
begin
  inherited;
  t:=FRoot.FindFile(FileName);
  Assert(Assigned(t));
  Stream.Size:=0;
  Stream.Write(t.FData^,t.FDataLength);
end;

procedure TVirtualFileSystem.RenameDirectory(OldPath, NewPath: string);
var
  t:TVirtualFolderItem;
  s1,s2:string;
begin
  inherited;
  t:=FRoot.FindFolder(OldPath);
  Assert(Assigned(t));
  Assert(not Assigned(FRoot.FindFolder(NewPath)));
  if NewPath[Length(NewPath)]='/' then
    NewPath:=Copy(NewPath,1,Length(NewPath)-1);
  SplitString(NewPath,s1,s2,['/'],True);
  t.SetName(s2);
end;

procedure TVirtualFileSystem.RenameFile(OldFileName, NewFileName: string);
var
  t:TVirtualFileItem;
  s1,s2:string;
begin
  inherited;
  t:=FRoot.FindFile(OldFileName);
  Assert(Assigned(t));
  Assert(not Assigned(FRoot.FindFile(NewFileName)));
  SplitString(NewFileName,s1,s2,['/'],True);
  t.SetName(s2);
end;

procedure TVirtualFileSystem.SaveVars(Vars: array of Pointer;
  Types: array of TDataType; Paths: array of string);
var
  a:Integer;
  s:TStream;
begin
  for a:=Low(Vars) to High(Vars) do begin
    case Types[a] of
      dtString:s:=TStringStream.Create(PString(Vars[a])^);
    else
      s:=TMemoryStream.Create;
      s.Write(Vars[a]^,GetDataSize(Types[a]));
    end;
    s.Position:=0;
    WriteStreamToFile(Paths[a],s);
    s.Destroy;
  end;
end;

procedure TVirtualFileSystem.WriteFileSystemToStream(s: TStream);
begin
  FRoot.WriteToStream(s);
end;

procedure TVirtualFileSystem.WriteStreamToFile(FileName: string;
  Stream: TStream);
var
  t:TVirtualFileItem;
  u:TVirtualFolderItem;
  s1,s2:string;
begin
  inherited;
  t:=FRoot.FindFile(FileName);
  if not Assigned(t) then begin
    SplitString(FileName,s1,s2,['/'],True);
    Assert(s2<>'');
    u:=FRoot.FindFolder(s1+'/');
    if not Assigned(u) then begin
      CreateDirectory(s1+'/');
      u:=FRoot.FindFolder(s1+'/');
    end;
    Assert(Assigned(u));
    t:=TVirtualFileItem.Create(u,s1+'/',s2,nil);
  end;
  t.SetData(Stream);
end;

{ TVirtualItem }

constructor TVirtualItem.Create(AParent: TVirtualFolderItem; APath, AName: string);
begin
  inherited Create;
  FParent:=AParent;
  if Assigned(FParent) then
    FParent.FSubItems.Add(Self);
  with FDescriptor do begin
    FileName:=AName;
    FilePath:=APath;
    FileTypeName:='';
    Folder:=False;
    Readable:=True;
    Writable:=True;
    FastExplore:=True;
    FileSize:=0;
    DateModified:=Now;
    OpenShellImageIndex:=-1;
    CloseShellImageIndex:=-1;
  end;
end;

destructor TVirtualItem.Destroy;
begin
  if Assigned(FParent) then
    FParent.FSubItems.Remove(Self);
  inherited;
end;

function ReadStringFromStream(s:TStream):string;
var
  a:Integer;
  p:PChar;
begin
  s.Read(a,SizeOf(a));
  GetMem(p,a+1);
  p[a]:=#0;
  s.Read(p^,a);
  Result:=p;
end;

procedure WriteStringToStream(Value:string;s:TStream);
var
  a:Integer;
  p:PChar;
begin
  p:=PChar(Value);
  a:=Length(Value);
  s.Write(a,SizeOf(a));
  s.Write(p^,a);
end;

procedure TVirtualItem.ReadFromStream(s: TStream);
begin
  with FDescriptor do begin
    FileName:=ReadStringFromStream(s);
    s.Read(DateModified,SizeOf(DateModified));
  end;
end;

procedure TVirtualItem.SetName(AName: string);
begin
  FDescriptor.FileName:=AName;
  FDescriptor.DateModified:=Now;
end;

procedure TVirtualItem.WriteToStream(s: TStream);
begin
  with FDescriptor do begin
    WriteStringToStream(FileName,s);
    s.Write(DateModified,SizeOf(DateModified));
  end;
end;

{ TVirtualFileItem }

constructor TVirtualFileItem.Create(AParent: TVirtualFolderItem; APath, AName: string; AData: TStream);
var
  s:string;
begin
  inherited Create(AParent,APath,AName);
  GetFileExtensionShellDescription('',s,FDescriptor.OpenShellImageIndex);
  FDescriptor.CloseShellImageIndex:=FDescriptor.OpenShellImageIndex;
  SetData(AData);
end;

destructor TVirtualFileItem.Destroy;
begin
  SetData(nil);
  inherited;
end;

function TVirtualFileItem.FindFile(Name: string): TVirtualFileItem;
begin
  if Name=FDescriptor.FileName then
    Result:=Self
  else
    Result:=nil;
end;

procedure TVirtualFileItem.ReadFromStream(s: TStream);
begin
  inherited;
  SetData(nil);
  s.Read(FDataLength,SizeOf(FDataLength));
  GetMem(FData,FDataLength);
  s.Read(FData^,FDataLength);
end;

procedure TVirtualFileItem.SetData(AData: TStream);
begin
  if Assigned(FData) then
    FreeMem(FData);
  if Assigned(AData) then begin
    FDataLength:=AData.Size;
    GetMem(FData,FDataLength);
    AData.Position:=0;
    AData.Read(FData^,FDataLength);
  end else begin
    FData:=nil;
    FDataLength:=0;
  end;
  FDescriptor.FileSize:=FDataLength;
  FDescriptor.DateModified:=Now;
end;

procedure TVirtualFileItem.WriteToStream(s: TStream);
begin
  inherited;
  s.Write(FDataLength,SizeOf(FDataLength));
  s.Write(FData^,FDataLength);
end;

{ TVirtualFolderItem }

constructor TVirtualFolderItem.Create(AParent: TVirtualFolderItem; APath, AName: string);
begin
  inherited;
  with FDescriptor do begin
    FileTypeName:='Folder';
    Folder:=True;
    OpenShellImageIndex:=DefaultFolderDescription.OpenImageIndex;
    CloseShellImageIndex:=DefaultFolderDescription.CloseImageIndex;
  end;
  FSubItems:=TObjectList.Create(False);
end;

destructor TVirtualFolderItem.Destroy;
begin
  while FSubItems.Count>0 do
    FSubItems[0].Destroy;
  FSubItems.Destroy;
  inherited;
end;

function TVirtualFolderItem.FindFile(Name: string): TVirtualFileItem;
var
  a:Integer;
  s1,s2:string;
  t:TVirtualFileItem;
begin
//  if Name[1]='/' then
//    Name:=Copy(Name,2,Length(Name)-1);
  Result:=nil;
  SplitString(Name,s1,s2,['/'],False);
  for a:=0 to FSubItems.Count-1 do
    with TVirtualItem(FSubItems[a]).FDescriptor do
      if Folder then begin
        if FileName=s1 then
          Result:=TVirtualFolderItem(FSubItems[a]).FindFile(s2)
      end else begin
        t:=TVirtualItem(FSubItems[a]).FindFile(s2);
        if Assigned(t) then
          Result:=t;
      end;
end;

function TVirtualFolderItem.FindFolder(Path: string): TVirtualFolderItem;
var
  a:Integer;
  s1,s2:string;
begin
  if ((Path='/') or (Path='//')) and (FDescriptor.FilePath+FDescriptor.FileName='') then begin
    Result:=Self;
    Exit;
  end;
  if (Path<>'') and (Path[1]='/') and (Path<>'/') then
    Path:=Copy(Path,2,Length(Path)-1);
  if Path='' then
    Result:=Self
  else begin
    Result:=nil;
    SplitString(Path,s1,s2,['/'],False);
    for a:=0 to FSubItems.Count-1 do
      with TVirtualItem(FSubItems[a]).FDescriptor do
        if Folder and (FileName=s1) then
          Result:=TVirtualFolderItem(FSubItems[a]).FindFolder(s2);
  end;
end;

procedure TVirtualFolderItem.ReadFromStream(s: TStream);
var
  a,b:Integer;
  t:Boolean;
  u:TVirtualItem;
begin
  inherited;
  s.Read(b,SizeOf(b));
  for a:=0 to b-1 do begin
    s.Read(t,SizeOf(t));
    if t then
      u:=TVirtualFolderItem.Create(Self,FDescriptor.FilePath+FDescriptor.FileName+'/','')
    else
      u:=TVirtualFileItem.Create(Self,FDescriptor.FilePath+FDescriptor.FileName+'/','',nil);
    u.ReadFromStream(s);
  end;
end;

procedure TVirtualFolderItem.WriteToStream(s: TStream);
var
  a,b:Integer;
  t:Boolean;
begin
  inherited;
  b:=FSubItems.Count;
  s.Write(b,SizeOf(b));
  for a:=0 to b-1 do begin
    t:=FSubItems[a] is TVirtualFolderItem;
    s.Write(t,SizeOf(t));
    TVirtualItem(FSubItems[a]).WriteToStream(s);
  end;

end;

end.
