unit ComponentCollection;

interface

uses
  SysUtils,Windows,Classes,Contnrs;

type
  TComponentCollection=class;

  IComponentCollectionOwner=interface
    ['{10A3B4C2-D3B9-4FFA-90A6-C66A418CC174}']
    function GetComponentCollection:TComponentCollection;
    property ComponentCollection:TComponentCollection read GetComponentCollection;
  end;

  IComponentCollectionNotifier=interface
    ['{8491B5D1-58CD-407A-B2FD-D3C44862DF1D}']
    procedure ReloadList(Sender:TComponentCollection);
  end;

  IComponentCollectionItem=interface
    ['{75166C37-DEE7-4B25-891C-85D5FBB49D6A}']
    function GetDesignIndex:Integer;
    function GetParent: TComponent;
    procedure SetParent(const Value: TComponent);
    property Parent:TComponent read GetParent write SetParent;
  end;

  TComponentCollectionItem=class(TComponent,IComponentCollectionItem)
  private
    FParent:TComponent;
    FIndex:Integer;

    function GetParent: TComponent;
    procedure SetParent(const Value: TComponent);
    function GetDesignIndex:Integer;
  protected
    procedure LoadIndex(Reader:TReader);
    procedure StoreIndex(Writer:TWriter);

    procedure DefineProperties(Filer:TFiler);override;
  public
    constructor Create(AOwner:TComponent);override;

    function GetParentComponent:TComponent;override;
    procedure SetParentComponent(Value:TComponent);override;
    function HasParent:Boolean;override;

    destructor Destroy;override;
  published
    property Parent:TComponent read GetParent write SetParent;
  end;

  TComponentDestructionEvent=procedure(AComponent:TComponent) of object;

  TComponentNotifier=class(TComponent)
  private
    FDestructionEvents:array of TComponentDestructionEvent;
  protected
    procedure Notification(AComponent:TComponent;Operation:TOperation);override;
  public
    constructor Create;reintroduce;

    procedure MonitorComponent(AComponent:TComponent);
    procedure UnMonitorComponent(AComponent:TComponent);

    procedure AddDestructionEvent(AComponentDestructionEvent:TComponentDestructionEvent);
    procedure RemoveDestructionEvent(AComponentDestructionEvent:TComponentDestructionEvent);

    destructor Destroy;override;
  end;

  TComponentCollection=class(TPersistent,IComponentCollectionOwner)
  private
    FOwner:TComponent;
    FList:TComponentList;
    FNotifiers:TComponentList;
    FDesigner:TComponent;
    FItemClass:TComponentClass;
    FModifiedCount:Integer;

    function QueryInterface(const IID:TGUID;out Obj):HResult;stdcall;
    function _AddRef:Integer;stdcall;
    function _Release:Integer;stdcall;

    procedure ComponentDestroyed(AComponent:TComponent);
  protected
    FComponentNotifier:TComponentNotifier;

    function GetOwner:TPersistent;override;

    procedure NotifyNotifiers;
  public
    constructor Create(AOwner:TComponent;AItemClass:TComponentClass);

    function GetComponentCollection:TComponentCollection;

    procedure BeforeModified;
    procedure AfterModified;
    procedure Modified;

    function Add(AComponent:TComponent):Integer;
    procedure Insert(AComponent:TComponent;Index:Integer);
    procedure Remove(AComponent:TComponent);
    function GetComponent(Index:Integer):TComponent;
    function IndexOf(AComponent:TComponent):Integer;
    procedure Exchange(x,y:Integer);

    function CanAccept(Value:TComponentClass):Boolean;virtual;

    function GetCount: Integer;
    property Count:Integer read GetCount;

    procedure AddNotifier(Value:TComponent);
    procedure RemoveNotifier(Value:TComponent);
    property Designer:TComponent read FDesigner write FDesigner;

    property Owner:TComponent read FOwner;

    destructor Destroy;override;
  published
  end;

  TComponentCollectionClass=class of TComponentCollection;
  TComponentCollectionItemClass=class of TComponentCollectionItem;

implementation

{ TComponentNotifier }

procedure TComponentNotifier.AddDestructionEvent(
  AComponentDestructionEvent: TComponentDestructionEvent);
begin
  SetLength(FDestructionEvents,high(FDestructionEvents)+2);
  FDestructionEvents[high(FDestructionEvents)]:=AComponentDestructionEvent;
end;

constructor TComponentNotifier.Create;
begin
  inherited Create(nil);
  SetLength(FDestructionEvents,0);
end;

destructor TComponentNotifier.Destroy;
begin
  SetLength(FDestructionEvents,0);
  inherited;
end;

procedure TComponentNotifier.MonitorComponent(AComponent: TComponent);
begin
  AComponent.FreeNotification(Self);
end;

procedure TComponentNotifier.Notification(AComponent: TComponent;
  Operation: TOperation);
var
  a:Integer;
  p:TComponentDestructionEvent;
begin
  inherited;
  p:=nil;
  if (Operation=opRemove) then begin
    for a:=high(FDestructionEvents) downto 0 do begin
      p:=FDestructionEvents[a];
      p(AComponent);
    end;
  end;
end;

procedure TComponentNotifier.RemoveDestructionEvent(
  AComponentDestructionEvent: TComponentDestructionEvent);
var
  a,b:Integer;
begin
  b:=-1;
  for a:=0 to high(FDestructionEvents) do
    if @FDestructionEvents=@AComponentDestructionEvent then
      b:=a;
  if b<>-1 then begin
    for a:=b to high(FDestructionEvents)-1 do
      FDestructionEvents[a]:=FDestructionEvents[a+1];
    SetLength(FDestructionEvents,high(FDestructionEvents));
  end;
end;

procedure TComponentNotifier.UnMonitorComponent(AComponent: TComponent);
begin
  AComponent.RemoveFreeNotification(Self);
end;

{ TComponentCollection }

function TComponentCollection.Add(AComponent: TComponent): Integer;
begin
  if not CanAccept(TComponentClass(AComponent.ClassType)) then
    raise Exception.Create('The collection cannot accept this component');
  FComponentNotifier.MonitorComponent(AComponent);
  Result:=FList.Add(AComponent);
  Modified;
end;

procedure TComponentCollection.AddNotifier(Value: TComponent);
begin
  if FNotifiers.IndexOf(Value)=-1 then
    FNotifiers.Add(Value);
end;

procedure TComponentCollection.AfterModified;
begin
  Dec(FModifiedCount);
  if FModifiedCount=0 then
    NotifyNotifiers;
end;

procedure TComponentCollection.BeforeModified;
begin
  Inc(FModifiedCount);
end;

function TComponentCollection.CanAccept(Value: TComponentClass): Boolean;
begin
  Result:=Value.InheritsFrom(FItemClass);
end;

procedure TComponentCollection.ComponentDestroyed(AComponent: TComponent);
begin
  Modified;
//  FComponentNotifier.UnmonitorComponent(AComponent);
end;

constructor TComponentCollection.Create(AOwner: TComponent; AItemClass: TComponentClass);
begin
  inherited Create;
  FOwner:=AOwner;
  FItemClass:=AItemClass;
  FComponentNotifier:=TComponentNotifier.Create;
  FComponentNotifier.AddDestructionEvent(ComponentDestroyed);
  FList:=TComponentList.Create(False);//True);
  FNotifiers:=TComponentList.Create(False);
end;

destructor TComponentCollection.Destroy;
begin
  if Assigned(FDesigner) then
    FDesigner.Destroy;
  FList.OwnsObjects:=True;
  FList.Destroy;
  FComponentNotifier.Destroy;
  FNotifiers.Destroy;
  inherited;
end;

procedure TComponentCollection.Exchange(x, y: Integer);
begin
  FList.Exchange(x,y);
  Modified;
end;

function TComponentCollection.GetComponent(Index: Integer): TComponent;
begin
  Result:=FList[Index];
end;

function TComponentCollection.GetComponentCollection: TComponentCollection;
begin
  Result:=Self;
end;

function TComponentCollection.GetCount: Integer;
begin
  Result:=FList.Count;
end;

function TComponentCollection.GetOwner: TPersistent;
begin
  Result:=FOwner;
end;

function TComponentCollection.IndexOf(AComponent: TComponent): Integer;
begin
  Result:=FList.IndexOf(AComponent);
end;

procedure TComponentCollection.Insert(AComponent: TComponent;
  Index: Integer);
begin
  FList.Insert(Index,AComponent);
  Modified;
end;

procedure TComponentCollection.Modified;
var
  a,b,c:Integer;
begin
  if csLoading in FOwner.ComponentState then begin
    for a:=0 to FList.Count-1 do begin
      c:=-1;
      for b:=0 to FList.Count-1 do
        if (FList[b] as IComponentCollectionItem).GetDesignIndex=a then begin
          c:=b;
          Break;
        end;
      if c=-1 then
        Exit;
      FList.Exchange(a,c);
    end;
  end;
  if FModifiedCount=0 then
    NotifyNotifiers;
end;

procedure TComponentCollection.NotifyNotifiers;
var
  a:Integer;
begin
  for a:=FNotifiers.Count-1 downto 0 do
    (FNotifiers[a] as IComponentCollectionNotifier).ReloadList(Self);
end;

function TComponentCollection.QueryInterface(const IID: TGUID;
  out Obj): HResult;
begin
  if GetInterface(IID,Obj) then
    Result:=0
  else
    Result:=E_NOINTERFACE;
end;

procedure TComponentCollection.Remove(AComponent: TComponent);
begin
  FList.Remove(AComponent)
end;

procedure TComponentCollection.RemoveNotifier(Value: TComponent);
begin

end;

function TComponentCollection._AddRef: Integer;
begin
  Result:=1;
end;

function TComponentCollection._Release: Integer;
begin
  Result:=1;
end;

{ TComponentCollectionItem }

constructor TComponentCollectionItem.Create(AOwner: TComponent);
begin
  inherited;
end;

procedure TComponentCollectionItem.DefineProperties(Filer: TFiler);
  function DoWrite:Boolean;
  begin
    Result:=True;
  end;
begin
  inherited;
  Filer.DefineProperty('Index',LoadIndex,StoreIndex,True);
end;

destructor TComponentCollectionItem.Destroy;
begin
  inherited;
end;

function TComponentCollectionItem.GetDesignIndex: Integer;
begin
  Result:=FIndex;
end;

function TComponentCollectionItem.GetParent: TComponent;
begin
  Result:=FParent;
end;

function TComponentCollectionItem.GetParentComponent: TComponent;
begin
  Result:=FParent;
end;

function TComponentCollectionItem.HasParent: Boolean;
begin
  Result:=False;//True;
end;

procedure TComponentCollectionItem.LoadIndex(Reader: TReader);
begin
  FIndex:=Reader.ReadInteger;
end;

procedure TComponentCollectionItem.SetParent(const Value: TComponent);
var
  j:IComponentCollectionOwner;
begin
  if Assigned(FParent) and ((FParent as IInterface).QueryInterface(IComponentCollectionOwner,j)=0) then
    j.ComponentCollection.Remove(Self);
  if Assigned(Value) and ((Value as IInterface).QueryInterface(IComponentCollectionOwner,j)=0) then
    j.ComponentCollection.Add(Self);
  FParent:=Value;
end;

procedure TComponentCollectionItem.SetParentComponent(Value: TComponent);
begin
  FParent:=Value;
end;

procedure TComponentCollectionItem.StoreIndex(Writer: TWriter);
var
  j:IComponentCollectionOwner;
  a:Integer;
begin
  a:=-1;
  if Assigned(FParent) and ((FParent as IInterface).QueryInterface(IComponentCollectionOwner,j)=0) then
    a:=j.ComponentCollection.IndexOf(Self);
  Writer.WriteVariant(a);
end;

end.
